/*
 * ixp1200.h -- CPU specific header for Cygnus BSP.
 *   Intel(R) IXP1200
 *
 * Copyright (c) 1999 Cygnus Support
 *
 * The authors hereby grant permission to use, copy, modify, distribute,
 * and license this software and its documentation for any purpose, provided
 * that existing copyright notices are retained in all copies and that this
 * notice is included verbatim in any distributions. No written agreement,
 * license, or royalty fee is required for any of the authorized uses.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * Intel is a Registered Trademark of Intel Corporation.
 * Other Brands and Trademarks are the property of their respective owners.
 */

#ifndef __IXP1200_H__
#define __IXP1200_H__ 1

#include <bsp/defs.h>

/*
 * Make sure the MMU is compiled in
 */
#define MMU

#ifdef __ASSEMBLER__

#define REG8_VAL(a)  (a)
#define REG16_VAL(a) (a)
#define REG32_VAL(a) (a)

#define REG8_PTR(a)  (a)
#define REG16_PTR(a) (a)
#define REG32_PTR(a) (a)

#else /* __ASSEMBLER__ */

#define REG8_VAL(a)  ((unsigned char)(a))
#define REG16_VAL(a) ((unsigned short)(a))
#define REG32_VAL(a) ((unsigned int)(a))

#define REG8_PTR(a)  ((volatile unsigned char *)(a))
#define REG16_PTR(a) ((volatile unsigned long *)(a))
#define REG32_PTR(a) ((volatile unsigned long *)(a))

#endif /* __ASSEMBLER__ */

/*
 * IXP1200 Cache and MMU Definitions
 */
#define IXP1200_ICACHE_SIZE                       SZ_16K
#define IXP1200_DCACHE_SIZE                       SZ_8K
#define IXP1200_ICACHE_LINESIZE_BYTES             32
#define IXP1200_DCACHE_LINESIZE_BYTES             32
#define IXP1200_ICACHE_LINESIZE_WORDS             8
#define IXP1200_DCACHE_LINESIZE_WORDS             8
#define IXP1200_ICACHE_WAYS                       32
#define IXP1200_DCACHE_WAYS                       32
#define IXP1200_ICACHE_LINE_BASE(p)               REG32_PTR((unsigned long)(p) & \
                                                           ~(IXP1200_ICACHE_LINESIZE_BYTES - 1))
#define IXP1200_DCACHE_LINE_BASE(p)               REG32_PTR((unsigned long)(p) & \
                                                           ~(IXP1200_DCACHE_LINESIZE_BYTES - 1))

#define IXP1200_ZEROS_BANK_BASE                (0x10000000)


/*
 * IXP1200 Control and Status Register Base Definitions
 */
/* 
 * We need 2 bases for registers, control&status and
 * peripheral.
 */

#define IXP1200_PERIPHERAL_BASE                0x90000000
#define IXP1200_CONTROL_STATUS_BASE              0x42000000

#define IXP1200_REGISTER(x)        REG32_PTR(IXP1200_CONTROL_STATUS_BASE + (x))
#define IXP1200_PERIPH_REGISTER(x) REG32_PTR(IXP1200_PERIPHERAL_BASE + (x))

/*
 * IXP1200 UART Control/Configuration Registers.
 */

#define IXP1200_UART_DATA_REGISTER_o               0x3C00
#define IXP1200_UART_RXSTAT_o                      0x3400
#define IXP1200_UART_H_BAUD_CONTROL_o              0x3800
#define IXP1200_UART_M_BAUD_CONTROL_o              0x3800
#define IXP1200_UART_L_BAUD_CONTROL_o              0x3800
#define IXP1200_UART_CONTROL_REGISTER_o            0x3800
#define IXP1200_UART_FLAG_REGISTER_o               0x3400


//	We added IXP1200_PERIPF_REGISTER for ixp1200
//	because the peripheral base is not the same as
//	control/status register base

#define IXP1200_UART_DATA_REGISTER \
               IXP1200_PERIPH_REGISTER(IXP1200_UART_DATA_REGISTER_o)
#define IXP1200_UART_RXSTAT \
               IXP1200_PERIPH_REGISTER(IXP1200_UART_RXSTAT_o)
#define IXP1200_UART_H_BAUD_CONTROL \
               IXP1200_PERIPH_REGISTER(IXP1200_UART_H_BAUD_CONTROL_o)
#define IXP1200_UART_M_BAUD_CONTROL \
               IXP1200_PERIPH_REGISTER(IXP1200_UART_M_BAUD_CONTROL_o)
#define IXP1200_UART_L_BAUD_CONTROL \
               IXP1200_PERIPH_REGISTER(IXP1200_UART_L_BAUD_CONTROL_o)
#define IXP1200_UART_CONTROL_REGISTER \
               IXP1200_PERIPH_REGISTER(IXP1200_UART_CONTROL_REGISTER_o)
#define IXP1200_UART_FLAG_REGISTER \
               IXP1200_PERIPH_REGISTER(IXP1200_UART_FLAG_REGISTER_o)

#define UART_BASE_0                    IXP1200_UART_DATA_REGISTER

/*
 * IXP1200 UART Data Register bit masks
 */
#define IXP1200_UART_DATA_MASK                     0x000000FF

/*
 * IXP1200 UART RX Status Register bit masks
 */
#define IXP1200_UART_FRAMING_ERROR_MASK            0x00000001
#define IXP1200_UART_PARITY_ERROR_MASK             0x00000002
#define IXP1200_UART_OVERRUN_ERROR_MASK            0x00000004

/*
 * IXP1200 UART High Baud Control Register bit masks
 */
#define IXP1200_UART_BREAK_DISABLED                0x00000000
#define IXP1200_UART_BREAK_ENABLED                 0x00000001
#define IXP1200_UART_BREAK_MASK                    0x00000001
#define IXP1200_UART_PARITY_DISABLED               0x00000000
#define IXP1200_UART_PARITY_ENABLED                0x00000002
#define IXP1200_UART_PARITY_MASK                   0x00000002
#define IXP1200_UART_PARITY_ODD                    0x00000000
#define IXP1200_UART_PARITY_EVEN                   0x00000004
#define IXP1200_UART_ODD_EVEN_SELECT_MASK          0x00000004
#define IXP1200_UART_STOP_BITS_ONE                 0x00000000
#define IXP1200_UART_STOP_BITS_TWO                 0x00000008
#define IXP1200_UART_STOP_BITS_SELECT_MASK         0x00000008
#define IXP1200_UART_FIFO_DISABLED                 0x00000000
#define IXP1200_UART_FIFO_ENABLED                  0x00000010
#define IXP1200_UART_FIFO_ENABLE_MASK              0x00000010
#define IXP1200_UART_DATA_LENGTH_5_BITS            0x00000000
#define IXP1200_UART_DATA_LENGTH_6_BITS            0x00000020
#define IXP1200_UART_DATA_LENGTH_7_BITS            0x00000040
#define IXP1200_UART_DATA_LENGTH_8_BITS            0x00000060
#define IXP1200_UART_DATA_LENGTH_MASK              0x00000060

/*
 * IXP1200 UART Medium Baud Control Register bit masks
 */
#define IXP1200_UART_H_BAUD_RATE_DIVISOR_MASK      0x0000000F

/*IXP 1200 */
#define UART_SR                         0x90003400
#define UART_CR                         0x90003800
#define UART_DR                         0x90003C00
/* UART Status Register Constants */
#define UART_TX_FIFO_FULL               BIT7
#define UART_RX_FIFO_FULL               BIT6
#define UART_TX_FIFO_EMPTY              BIT5
#define UART_RX_READY                   BIT4
#define UART_RX_OVERRUN                 BIT3
#define UART_TX_READY                   BIT2
#define UART_FRAMING_ERR                BIT1
#define UART_PARITY_ERR                 BIT0

/* UART Control Register Contants */
#define UART_INT_CLEAR                  BIT9
#define UART_TX_INT_ENABLE              BIT8
#define UART_ENABLE_BIT                 BIT7
#define UART_DATA_BITS_8                (BIT6 | BIT5)
#define UART_RX_INT_ENABLE              BIT4
#define UART_STOP_BIT_1                 BIT3
#define UART_PARITY_ENABLE              BIT1
#define UART_BREAK                      BIT0
/*IXP 1200 */
/*
 * IXP1200 UART Low Baud Control Register bit masks
 */
#define IXP1200_UART_L_BAUD_RATE_DIVISOR_MASK      0x000000FF

/*
 * IXP1200 UART Control Register bit fields
 */
#define IXP1200_UART_DISABLED			0x00000000
#define IXP1200_UART_ENABLED                    0x00000001
#define IXP1200_UART_ENABLE_MASK                0x00000001
#define IXP1200_SIR_DISABLED                    0x00000000

/*
 * IXP1200 IRQ Controller Registers
 */
#define IXP1200_IRQCONT_IRQSTATUS_o                0x180
#define IXP1200_IRQCONT_IRQRAWSTATUS_o             0x184
#define IXP1200_IRQCONT_IRQENABLE_o                0x188
#define IXP1200_IRQCONT_IRQENABLESET_o             0x188
#define IXP1200_IRQCONT_IRQENABLECLEAR_o           0x18C
#define IXP1200_IRQCONT_IRQSOFT_o                  0x190
#define IXP1200_IRQCONT_FIQSTATUS_o                0x280
#define IXP1200_IRQCONT_FIQRAWSTATUS_o             0x284
#define IXP1200_IRQCONT_FIQENABLE_o                0x288
#define IXP1200_IRQCONT_FIQENABLESET_o             0x288
#define IXP1200_IRQCONT_FIQENABLECLEAR_o           0x28C
#define IXP1200_IRQCONT_FIQSOFT_o                  0x290

/*
 * IXP1200 IRQ Controller Registers
 */
#define IXP1200_IRQCONT_IRQSTATUS_o                0x180
#define IXP1200_IRQCONT_IRQRAWSTATUS_o             0x184
#define IXP1200_IRQCONT_IRQENABLE_o                0x188
#define IXP1200_IRQCONT_IRQENABLESET_o             0x188
#define IXP1200_IRQCONT_IRQENABLECLEAR_o           0x18C
#define IXP1200_IRQCONT_IRQSOFT_o                  0x190
#define IXP1200_IRQCONT_FIQSTATUS_o                0x280
#define IXP1200_IRQCONT_FIQRAWSTATUS_o             0x284
#define IXP1200_IRQCONT_FIQENABLE_o                0x288
#define IXP1200_IRQCONT_FIQENABLESET_o             0x288
#define IXP1200_IRQCONT_FIQENABLECLEAR_o           0x28C
#define IXP1200_IRQCONT_FIQSOFT_o                  0x290

#define IXP1200_IRQCONT_IRQSTATUS                  IXP1200_REGISTER(IXP1200_IRQCONT_IRQSTATUS_o)
#define IXP1200_IRQCONT_IRQRAWSTATUS               IXP1200_REGISTER(IXP1200_IRQCONT_IRQRAWSTATUS_o)
#define IXP1200_IRQCONT_IRQENABLE                  IXP1200_REGISTER(IXP1200_IRQCONT_IRQENABLE_o)
#define IXP1200_IRQCONT_IRQENABLESET               IXP1200_REGISTER(IXP1200_IRQCONT_IRQENABLESET_o)
#define IXP1200_IRQCONT_IRQENABLECLEAR             IXP1200_REGISTER(IXP1200_IRQCONT_IRQENABLECLEAR_o)
#define IXP1200_IRQCONT_IRQSOFT                    IXP1200_REGISTER(IXP1200_IRQCONT_IRQSOFT_o)
#define IXP1200_IRQCONT_FIQSTATUS                  IXP1200_REGISTER(IXP1200_IRQCONT_FIQSTATUS_o)
#define IXP1200_IRQCONT_FIQRAWSTATUS               IXP1200_REGISTER(IXP1200_IRQCONT_FIQRAWSTATUS_o)
#define IXP1200_IRQCONT_FIQENABLE                  IXP1200_REGISTER(IXP1200_IRQCONT_FIQENABLE_o)
#define IXP1200_IRQCONT_FIQENABLESET               IXP1200_REGISTER(IXP1200_IRQCONT_FIQENABLESET_o)
#define IXP1200_IRQCONT_FIQENABLECLEAR             IXP1200_REGISTER(IXP1200_IRQCONT_FIQENABLECLEAR_o)
#define IXP1200_IRQCONT_FIQSOFT                    IXP1200_REGISTER(IXP1200_IRQCONT_FIQSOFT_o)

/*
 * IRQ Controller IRQ Numbers
 */
#define IXP1200_IRQ_MIN                            0

#define IXP1200_IRQ_RSV0                           0
#define IXP1200_IRQ_SOFT_INTERRUPT                 1
#define IXP1200_IRQ_CONSOLE_RX                     2
#define IXP1200_IRQ_CONSOLE_TX                     3
#define IXP1200_IRQ_TIMER_1                        4
#define IXP1200_IRQ_TIMER_2                        5
#define IXP1200_IRQ_TIMER_3                        6
#define IXP1200_IRQ_TIMER_4                        7
#define IXP1200_IRQ_IRQ_IN_I_0                     8
#define IXP1200_IRQ_IRQ_IN_I_1                     9
#define IXP1200_IRQ_IRQ_IN_I_2                     10
#define IXP1200_IRQ_IRQ_IN_I_3                     11
#define IXP1200_IRQ_XCS_I_0                        12
#define IXP1200_IRQ_XCS_I_1                        13
#define IXP1200_IRQ_XCS_I_2                        14
#define IXP1200_IRQ_DOORBELL_FROM_HOST             15
#define IXP1200_IRQ_DMA_CHAN_1                     16
#define IXP1200_IRQ_DMA_CHAN_2                     17
#define IXP1200_IRQ_PIC_IRQ_I                      18
#define IXP1200_IRQ_PMCSR_WRITE_BY_HOST            19
#define IXP1200_IRQ_RSV1                           20
#define IXP1200_IRQ_RSV2                           21
#define IXP1200_IRQ_START_BIST                     22
#define IXP1200_IRQ_RECEIVED_SERR                  23
#define IXP1200_IRQ_SDRAM_PARITY                   24
#define IXP1200_IRQ_I20_INBOUND_POST_LIST          25
#define IXP1200_IRQ_RSV3                           26
#define IXP1200_IRQ_DISCARD_TIMER_EXPIRED          27
#define IXP1200_IRQ_DATA_PARITY_ERROR              28
#define IXP1200_IRQ_MASTER_ABORT                   29
#define IXP1200_IRQ_TARGET_ABORT                   30
#define IXP1200_IRQ_PARITY_ERROR                   31

#define IXP1200_IRQ_MAX                            31
#define NUM_IXP1200_INTERRUPTS                     IXP1200_IRQ_MAX - IXP1200_IRQ_MIN + 1
#define IXP1200_IRQ_INTSRC_MASK(irq_nr)            (1 << (irq_nr))


/*
 * IXP1200 IRQSOFT/FIQSOFT Register bit fields
 */
#define IXP1200_IRQSOFT_RAW_BIT_MASK               0x00000001

#endif /* __IXP1200_H__ */
