/*
 * ixp-1200.c -- Intel(R) IXP1200 specific support for Cygnus BSP
 *
 * Copyright (c) 1998, 1999 Cygnus Support
 *
 * The authors hereby grant permission to use, copy, modify, distribute,
 * and license this software and its documentation for any purpose, provided
 * that existing copyright notices are retained in all copies and that this
 * notice is included verbatim in any distributions. No written agreement,
 * license, or royalty fee is required for any of the authorized uses.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * Intel is a Registered Trademark of Intel Corporation.
 * ARM is a Registered Trademark of Advanced RISC Machines Limited.
 * Other Brands and Trademarks are the property of their respective owners.
 */

#include <stdlib.h>
#include <bsp/ixp1200.h>
#include <bsp/cpu.h>
#include <bsp/bsp.h>
#include "bsp_if.h"
#include "../include/debflags.h"
#include "queue.h"

#define UART_SPEED_AT_9600	1
/*PVU*/
#define UART_DR_BASE ((volatile unsigned int *)UART_DR)
#define UART_CR_BASE ((volatile unsigned int *)UART_CR)
#define UART_SR_BASE ((volatile unsigned int *)UART_SR)

//#define CHIP_REV_BEFORE_B0	1

#define USE_UART_INTS      1

#ifdef CHIP_REV_BEFORE_B0
#define CLR_UART_INTS_BUG    (*UART_CR_BASE |= UTCR_INT_CLEAR)
#endif //#ifdef CHIP_REV_BEFORE_B0

void write_led (unsigned char);
static void uart_irq_enable(void *base);
static int uart_irq_disable(void *base);

#define UART_CLOCK                      3686400
//#define UART_CLOCK                      100000000/32


#define IRQ_UART_TX_MASK    (1 << 8      )
#define IRQ_UART_RX_MASK    (1 << 4      )

#define IRQ_REG	        REG32_PTR(0x90001400)

#ifdef CHIP_REV_BEFORE_B0
#define IRQ_UART_INT	0x40	// UART Interrupt for chip rev. before B0
#else //#ifdef CHIP_REV_BEFORE_B0
#define IRQ_UART_INT	0x100	// UART Interrupt for chip rev. B0 and beyond
#endif //#ifdef CHIP_REV_BEFORE_B0

//         Setting UTCR_INT_CLEAR bit in interrupt handler
//         is a temporary solution. This should not be necessary
//         in future revisions of ixp1200 board
#ifdef CHIP_REV_BEFORE_B0
#define UTCR_INT_CLEAR	0x200		// UART Interrupt Clear
#endif //#ifdef CHIP_REV_BEFORE_B0

#define BAUD_9600                       (((UART_CLOCK/16/9600)-1)   << 16)
/*PVU*/

#define DUMP_CTL_STAT_REGS \
    bsp_printf("*IXP1200_IRQCONT_IRQSTATUS    = 0x%08lx, addr=0x%lx\n", \
        *IXP1200_IRQCONT_IRQSTATUS, IXP1200_IRQCONT_IRQSTATUS); \
    bsp_printf("*IXP1200_IRQCONT_IRQRAWSTATUS = 0x%08lx, addr=0x%lx\n", \
        *IXP1200_IRQCONT_IRQRAWSTATUS, IXP1200_IRQCONT_IRQRAWSTATUS); \
    bsp_printf("*IXP1200_IRQCONT_IRQENABLE    = 0x%08lx, addr=0x%lx\n", \
        *IXP1200_IRQCONT_IRQENABLE, IXP1200_IRQCONT_IRQENABLE); \
    bsp_printf("*IXP1200_IRQCONT_FIQSTATUS    = 0x%08lx, addr=0x%lx\n", \
        *IXP1200_IRQCONT_FIQSTATUS, IXP1200_IRQCONT_FIQSTATUS); \
    bsp_printf("*IXP1200_IRQCONT_FIQRAWSTATUS = 0x%08lx, addr=0x%lx\n", \
        *IXP1200_IRQCONT_FIQRAWSTATUS, IXP1200_IRQCONT_FIQRAWSTATUS); \
    bsp_printf("*IXP1200_IRQCONT_FIQENABLE    = 0x%08lx, addr=0x%lx\n", \
        *IXP1200_IRQCONT_FIQENABLE, IXP1200_IRQCONT_FIQENABLE); \
    bsp_printf("*UART_CR_BASE               = 0x%08lx, addr=0x%lx\n", \
        *UART_CR_BASE, UART_CR_BASE); \
    bsp_printf("*IRQ_REG                    = 0x%08lx, addr=0x%lx\n", \
        *IRQ_REG, IRQ_REG);


/*
 * Enable/Disable separate debug functions
 */

static void	ixp1200_ictrl_init(const struct bsp_irq_controller *ic);
static int	ixp1200_ictrl_disable(const struct bsp_irq_controller *ic, int irq_nr);
static void	ixp1200_ictrl_enable(const struct bsp_irq_controller *ic, int irq_nr);

/* static bsp_vec_t *ixp1200_irq_vec_tbl[NUM_IXP1200_INTERRUPTS]; */ 
bsp_vec_t *ixp1200_irq_vec_tbl[NUM_IXP1200_INTERRUPTS];

static const struct bsp_irq_controller ixp1200_irq_controller = {
    IXP1200_IRQ_MIN, IXP1200_IRQ_MAX,
    &ixp1200_irq_vec_tbl[0],
    ixp1200_ictrl_init,
    ixp1200_ictrl_disable,
    ixp1200_ictrl_enable
};

void write_led(unsigned char ch)
{
	unsigned char *led;
	
	led = (unsigned char *) 0x38508000;
	*led = ch;
}

/*
 * Dispatch code for ixp1200 interrupts. Called by exception dispatch code.
 */
static int
ixp1200_irq_dispatch(int ex_nr, ex_regs_t *regs)
{
    bsp_vec_t *vec = NULL;
    int done = 0;
    int irq = 0;

/*
    write_led (0x70);
*/

#if  DEBUG_IRQ
    bsp_printf("ixp1200_irq_dispatch() exc_nr %d regs %lx\n",
                 ex_nr, regs);
#endif // DEBUG_IRQ

#if DEBUG_UART_CODE
    BSP_ASSERT((ex_nr == BSP_CORE_EXC_IRQ) || (ex_nr == BSP_CORE_EXC_FIQ));

    DUMP_CTL_STAT_REGS
#endif /* DEBUG_UART_CODE */

    if (ex_nr == BSP_CORE_EXC_FIQ)
    {
        /*
         * Some FIQ occurred.
         * Figure out which one.
         */
        unsigned long int_status = *IXP1200_IRQCONT_FIQSTATUS;

        /*
         * We will prioritize them by the low-order bit that is
         * set in the status register
         */
        while (((int_status & 0x1) == 0x0) && (irq < NUM_IXP1200_INTERRUPTS))
        {
            irq++;
            int_status >>= 1;
        }
    } else {
        /*
         * Some IRQ occurred.
         * Figure out which one.
         */
        if (*IRQ_REG & IRQ_UART_INT)
            irq = IXP1200_IRQ_CONSOLE_RX;    // UART RX interrupt (=2 on IXP1200)
        else
        {
            unsigned long int_status = *IXP1200_IRQCONT_IRQSTATUS;

            /*
             * We will prioritize them by the low-order bit that is
             * set in the status register
             */
            while (((int_status & 0x1) == 0x0) && (irq < NUM_IXP1200_INTERRUPTS))
            {
                irq++;
                int_status >>= 1;
            }
        }
    }

#if DEBUG_UART_CODE
    bsp_printf("irq = %d\n", irq);
#endif /* DEBUG_UART_CODE */

    if (irq < NUM_IXP1200_INTERRUPTS)
        vec = ixp1200_irq_vec_tbl[irq];

    while (!done && vec && vec->handler != NULL) {
#if DEBUG_UART_CODE
        bsp_printf("ixp1200_irq_dispatch() vec->handler=0x%08lx, vec->next=%lx\n",
		vec->handler, vec->next);
#endif /* DEBUG_UART_CODE */
        done = vec->handler(ex_nr, regs);
        vec = vec->next;
    }

    return done;
}


/*
 *  Initialize IXP1200 interrupt controller.
 */
static void
ixp1200_ictrl_init(const struct bsp_irq_controller *ic)
{
    static bsp_vec_t irq_vec, fiq_vec;
    int i;

#if DEBUG_UART_CODE
    bsp_printf("ixp1200_ictrl_init(0x%08lx)\n", ic);
    DUMP_CTL_STAT_REGS
#endif /* DEBUG_UART_CODE */

    /*
     * Clear all pending interrupts
     */
    *IXP1200_IRQCONT_IRQENABLECLEAR = ~0;
    *IXP1200_IRQCONT_FIQENABLECLEAR = ~0;

    for (i = ic->first; i < ic->last; i++)
    {
        ic->vec_list[i - ic->first] = NULL;
    }

    /*
     * Setup exception handlers for the exceptions
     * corresponding to IRQ's and FIQ's
     */
    irq_vec.handler = (void *)ixp1200_irq_dispatch;
    irq_vec.next = NULL;

#if DEBUG_IRQ
    bsp_printf("ixp1200_ictrl_init() inst hndlr BSP_CORE_EXC_IRQ %lx\n",
		 ixp1200_irq_dispatch);
#endif // DEBUG_IRQ

    bsp_install_vec(BSP_VEC_EXCEPTION, BSP_CORE_EXC_IRQ, BSP_VEC_CHAIN_FIRST,
			&irq_vec);

    fiq_vec.handler = (void *)ixp1200_irq_dispatch;
    fiq_vec.next = NULL;

#if DEBUG_IRQ
    bsp_printf("ixp1200_ictrl_init() inst hndlr BSP_CORE_EXC_FIQ %lx\n",
                 ixp1200_irq_dispatch);
#endif // DEBUG_IRQ

    bsp_install_vec(BSP_VEC_EXCEPTION, BSP_CORE_EXC_FIQ, BSP_VEC_CHAIN_FIRST,
			&fiq_vec);
}


/*
 *  Disable IXP1200 interrupts.
 */
static int
ixp1200_ictrl_disable(const struct bsp_irq_controller *ic, int irq_nr)
{
#if DEBUG_UART_CODE
    bsp_printf("ixp1200_ictrl_disable(0x%08lx, %d)\n", ic, irq_nr);
    BSP_ASSERT((irq_nr >= 0) && (irq_nr <= NUM_IXP1200_INTERRUPTS));
#endif /* DEBUG_UART_CODE */

    if (irq_nr == IXP1200_IRQ_CONSOLE_RX)
    {
        uart_irq_disable((void *) UART_CR_BASE);
#if DEBUG_UART_CODE
        bsp_printf("ixp1200_ictrl_disable() -> uart_irq_disable()\n");
#endif /* DEBUG_UART_CODE */
    }
    else
        *IXP1200_IRQCONT_IRQENABLECLEAR = IXP1200_IRQ_INTSRC_MASK(irq_nr);

#if DEBUG_UART_CODE
    DUMP_CTL_STAT_REGS
#endif /* DEBUG_UART_CODE */

    return 0;
}


/*
 *  Enable IXP1200 interrupts.
 */
static void
ixp1200_ictrl_enable(const struct bsp_irq_controller *ic, int irq_nr)
{

#if DEBUG_UART_CODE
    bsp_printf("ixp1200_ictrl_enable(0x%08lx, %d)\n", ic, irq_nr);
    BSP_ASSERT((irq_nr >= 0) && (irq_nr <= NUM_IXP1200_INTERRUPTS));
#endif /* DEBUG_UART_CODE */

    if (irq_nr == IXP1200_IRQ_CONSOLE_RX)
    {
        uart_irq_enable((void *) UART_CR_BASE);
#if DEBUG_UART_CODE
        bsp_printf("ixp1200_ictrl_enable() -> uart_irq_enable()\n");
#endif /* DEBUG_UART_CODE */
    }
    else
        *IXP1200_IRQCONT_IRQENABLESET = IXP1200_IRQ_INTSRC_MASK(irq_nr);

#if DEBUG_UART_CODE
    DUMP_CTL_STAT_REGS
#endif /* DEBUG_UART_CODE */
}


/*
 * IXP1200 specific routine to 'install' interrupt controllers.
 */
void
_bsp_install_ixp1200_irq_controllers(void)
{
#if DEBUG_UART_CODE
    bsp_printf("_bsp_install_ixp1200_irq_controllers()\n");
#endif /* DEBUG_UART_CODE */

    _bsp_install_irq_controller(&ixp1200_irq_controller);
}

/*
 * Function prototypes
 */
static void uart_write   (void*, const char*, int);
static int  uart_read    (void*, char*, int);
static void uart_putchar (void*, char);
static int  uart_getchar (void*);
static int  uart_control (void*, int, ...);
#if DEBUG_UART_FUNCTIONS
void uart_putnibble(void*, unsigned char);
void uart_putbyte(void*, unsigned char);
void uart_putword(void*, unsigned short);
void uart_putlong(void*, unsigned long);
void uart_putstring(void *, char *);
void uart_putcharacter(void *, char);
#endif /* DEBUG_UART_FUNCTIONS */

/*
 * Setup the bsp_comm_channel data structure
 */
#define BSP_NUM_COMMS        1
struct bsp_comm_channel _bsp_comm_list[BSP_NUM_COMMS] = 
{
    {
	{ "Serial Port A", BSP_COMM_SERIAL, BSP_PROTO_NONE },
	{ (void*)UART_DR_BASE /*UART_BASE_0*/, uart_write, uart_read,
	  uart_putchar, uart_getchar, uart_control }
    }
};
int _bsp_num_comms = sizeof(_bsp_comm_list)/sizeof(_bsp_comm_list[0]); 

/*
 * Baud rate selection stuff
 */
static int uart_baud[BSP_NUM_COMMS] = {-1};
struct _baud {
    int baud;
    unsigned short divisor_high, divisor_low;
};

#ifndef FCLK_MHZ
#error MUST DEFINE FCLK_MHZ FOR IXP1200 BASED BOARDS
#endif /* FCLK_MHZ */

const static struct _baud bauds[] = {
#if (FCLK_MHZ == 50)
    {   300, 0xA, 0x2B},                  /* 2603  = 0x0A2B */
    {   600, 0x5, 0x15},                  /* 1301  = 0x0515 */
    {  1200, 0x2, 0x8A},                  /* 650   = 0x028A */
    {  2400, 0x1, 0x45},                  /* 325   = 0x0145 */
    {  4800, 0x0, 0xA2},                  /* 162   = 0x00A2 */
    {  9600, 0x0, 0x50},                  /* 80    = 0x0050 */
    { 19200, 0x0, 0x28},                  /* 40    = 0x0028 */
    { 38400, 0x0, 0x13},                  /* 19    = 0x0013 */
#elif (FCLK_MHZ == 60)
    {   300, 0xC, 0x34},                  /* 2603  = 0x0A2B */
    {   600, 0x6, 0x19},                  /* 1301  = 0x0515 */
    {  1200, 0x3, 0x0C},                  /* 650   = 0x028A */
    {  2400, 0x1, 0x86},                  /* 325   = 0x0145 */
    {  4800, 0x0, 0xC2},                  /* 162   = 0x00A2 */
    {  9600, 0x0, 0x61},                  /* 80    = 0x0050 */
    { 19200, 0x0, 0x30},                  /* 40    = 0x0028 */
    { 38400, 0x0, 0x17},                  /* 19    = 0x0013 */
#endif
};

/*
 * UART queuing stuff
 */
#define UART_BUFSIZE 256
static bsp_queue_t uart_queue[BSP_NUM_COMMS];
static char uart_buffer[BSP_NUM_COMMS][UART_BUFSIZE];

/*
 * The currently selected debug port
 */
static volatile unsigned char *uart_dbg_base = NULL;

static void
uart_irq_enable(void *base)
{

#if 0	// Enabling this print statement produces this message
	// every time the key is hit.
#ifdef  DEBUG_UART_CODE
        bsp_printf("uart_irq_enable(), base %lx\n", UART_CR_BASE);
#endif // DEBUG_UART_CODE
#endif // #if 0

    *UART_CR_BASE |= IRQ_UART_RX_MASK;
}


//
// ------------------------------------------------
// Disable UART IRQ.
// Return:
//    Previous enabled/disabled flag
// ------------------------------------------------
//
static int
uart_irq_disable(void *base)
{
    int was_enabled;

    was_enabled = *UART_CR_BASE & IRQ_UART_RX_MASK;

#if 0
#ifdef  DEBUG_UART_CODE
        bsp_printf("uart_irq_disable(), was=0x%08lx, base 0x%lx\n",
		was_enabled, UART_CR_BASE);
#endif // DEBUG_UART_CODE
#endif // #if 0

    *UART_CR_BASE &= ~IRQ_UART_RX_MASK;
    return was_enabled;
}

/*static ORIG*/ 
void uart_putchar(void *base, char c)
{
	while (((*UART_SR_BASE) & UART_TX_READY )== 0)
	    ; 

	*UART_DR_BASE = c;
}


static void
uart_write(void *base, const char *buf, int len)
{
    while (len-- > 0)
        uart_putchar(base, *buf++);
}

int
uart_rcvchar(void *base)
{
    int result_char;

	while (((*UART_SR_BASE) & UART_RX_READY )== 0)
	    ; 

	result_char = (*UART_DR_BASE & 0x00FF); 

//         We are clearing the UART interrupts here
//         because of a chipset bug. This should be taken care of
//         in future versions of the chipset.

#ifdef CHIP_REV_BEFORE_B0
    	CLR_UART_INTS_BUG;
#endif //#ifdef CHIP_REV_BEFORE_B0

	return(result_char); 
}

static int
uart_getchar(void *base)
{
    int c, was_enabled;

#if USE_UART_INTS
    was_enabled = uart_irq_disable(base);
    c = _bsp_dequeue(&uart_queue[0]);

#if DEBUG_UART_CHAR
    bsp_printf("uart_getchar(), dequed char=%d\n", c);
#endif // DEBUG_UART_CHAR

    if (c < 0)
#endif // USE_UART_INTS

        c = uart_rcvchar(base);

#if USE_UART_INTS
    if (was_enabled)
	uart_irq_enable(base);
#endif // USE_UART_INTS

#if DEBUG_LOCAL_ECHO
    uart_putchar(base, c);
#endif // USE_UART_INTS

    return c;
}

static int
uart_read(void *base, char *buf, int len)
{
    int ch;

    if (len <= 0)
        return 0;

    ch = uart_getchar(base);
    *buf = ch;
    return 1;
}

static int
uart_irq_handler(int irq_nr, void *regs)
{
    unsigned char ch;
    int return_code = 0;

#if DEBUG_UART_CODE
    bsp_printf("In uart_irq_handler(%d, 0x%08lx)\n", irq_nr, regs);
    _bsp_dump_regs(regs);
#endif /* DEBUG_UART_CODE */

    if (*UART_CR_BASE & IRQ_UART_RX_MASK)
    {
#if DEBUG_UART_CODE
           bsp_printf("uart_irq_handler() IRQ_UART_RX_MASK\n");
#endif /* DEBUG_UART_CODE */

	if ((*UART_SR_BASE) & UART_RX_READY )
        {
#if DEBUG_UART_CODE
            bsp_printf("uart_irq_handler() UART_RX_READY\n");
#endif /* DEBUG_UART_CODE */

            ch = (*UART_DR_BASE & 0x00FF); 

	    /*
	     * If we read a Ctrl-C then we need to let
	     * GDB handle it.  Return 0 to have the GDB
	     * exception processor run.
	     */
	    if (ch == '\003')
	    {
#if DEBUG_GDB
                bsp_printf("uart_irq_handler() got ^C, must be GDB\n");
#endif // DEBUG_GDB
		return_code = 0;
	    } else {
		_bsp_enqueue(&uart_queue[0], ch);
#if DEBUG_LOCAL_ECHO
		uart_putchar(uart_dbg_base, ch);
#endif
		return_code = 1;
	    }
	}
    }
    else
	return_code = 1;
  
#ifdef CHIP_REV_BEFORE_B0
    CLR_UART_INTS_BUG;
#endif //#ifdef CHIP_REV_BEFORE_B0

// NOTE:   If we get into this routine, but ther is no character
//         available, we will return 0, meaning interrupt NOT handled.
//         In this case, cygmon takes control because of unhandled
//         exception, stopping any work in progress.
//         By clearing the interrupts in uart_rcvchar(),
//         using CLR_UART_INTS_BUG define, we no longer need
//         to force the return_code = 1 here.
//
//    return_code = 1;

#if DEBUG_UART_CODE
    bsp_printf("uart_irq_handler() returning %d\n", return_code);
#endif /* DEBUG_UART_CODE */

    return return_code;
}

static int
uart_setbaud(void *base, int bps)
{
	unsigned int baud;

	baud = (((UART_CLOCK/16)/bps) - 1)<<16;

#if USE_UART_INTS 
//         We disable the UART_RX_INT_ENABLE RX interrupt setting here,
//         as it is done explicitly by irq enable routine.

    *UART_CR_BASE = baud        | \
                    UART_DATA_BITS_8 | UART_ENABLE_BIT
//                    | UART_TX_INT_ENABLE
//                    | UART_RX_INT_ENABLE
			;
#else
    *UART_CR_BASE = baud        | \
                    UART_DATA_BITS_8 |UART_ENABLE_BIT;
#endif // #if USE_UART_INTS 

#if 0
    if (uart_baud[0] != bps)
    {
        for (i = 0, bp = bauds; i < (sizeof(bauds)/sizeof(bauds[0])); i++, bp++)
        {
            if (bp->baud == bps)
            {
                /*
                 * Setup the UART Baud Rate Divisor
                 *
                 * Note that the ordering of these writes is critical,
                 * and the writes to the H_BAUD_CONTROL and CONTROL_REGISTER
                 * are necessary to force the UART to update its register
                 * contents.
                 */
                *IXP1200_UART_L_BAUD_CONTROL   = bp->divisor_low;
                *IXP1200_UART_M_BAUD_CONTROL   = bp->divisor_high;
                *IXP1200_UART_H_BAUD_CONTROL = IXP1200_UART_BREAK_DISABLED    |
                                             IXP1200_UART_PARITY_DISABLED   |
                                             IXP1200_UART_STOP_BITS_ONE     |
                                             IXP1200_UART_FIFO_ENABLED      |
                                             IXP1200_UART_DATA_LENGTH_8_BITS;
                *IXP1200_UART_CONTROL_REGISTER = IXP1200_UART_ENABLED | 
                                               IXP1200_SIR_DISABLED;


                /*
                 * Store the state
                 */
                uart_baud[0] = bps;
                return 0;
            }
        }
    }

    return -1;
#endif // #if 0

//         There was no return value here because it was ifdefed out
//         We return 0 to indicate success
    return 0;
}

static int
uart_control(void *base, int func, ...)
{
    int                    i, retval = 0;
    va_list                ap;
    static bsp_vec_t       uart_vec;

    va_start (ap, func);

#if DEBUG_IRQ
    bsp_printf("uart_control() base %lx func %d\n",
                 base, func);
#endif // DEBUG_IRQ

    switch (func) {
    case COMMCTL_SETBAUD:
        i = va_arg(ap, int);
        retval = uart_setbaud((void *)base, i);
        break;

    case COMMCTL_GETBAUD:
        retval = uart_baud[0];
        break;

    case COMMCTL_INSTALL_DBG_ISR:
        if (uart_dbg_base != base) {
            if (uart_dbg_base == NULL) {

                uart_vec.handler = uart_irq_handler;
                uart_vec.next = NULL;

#if DEBUG_IRQ
        bsp_printf("uart_control() -> bsp_install_vec(), handlr %lx\n",
                 uart_irq_handler);
#endif // DEBUG_IRQ

                bsp_install_vec(BSP_VEC_INTERRUPT, IXP1200_IRQ_CONSOLE_RX,
                                BSP_VEC_REPLACE, &uart_vec);
                bsp_enable_irq(IXP1200_IRQ_CONSOLE_RX);
                uart_irq_enable(base);
            }
            uart_dbg_base = base;
        }
        break;
	
    case COMMCTL_REMOVE_DBG_ISR:
        uart_irq_disable(base);
        bsp_disable_irq(IXP1200_IRQ_CONSOLE_RX);
        bsp_remove_vec(BSP_VEC_INTERRUPT, IXP1200_IRQ_CONSOLE_RX, &uart_vec);
        uart_dbg_base = NULL;
        break;

    case COMMCTL_IRQ_DISABLE:
	retval = uart_irq_disable(base);
	break;

    case COMMCTL_IRQ_ENABLE:
	uart_irq_enable(base);
	break;

    default:
        retval = -1;
        break;
    }

    va_end(ap);

    return retval;
}


#if DEBUG_UART_FUNCTIONS
#else
static
#endif /* DEBUG_UART_FUNCTIONS */
void
init_uart(void *base, int bps)
{
    unsigned dummy;

    *UART_CR_BASE = 0;	// Totally reset uart

    /*
     * Read the RXStat to drain the fifo
     */
    dummy = *IXP1200_UART_RXSTAT;

    /*
     * Set the baud rate - this also turns the uart on.
     */
    uart_setbaud((void *)UART_CR_BASE, bps);

}

/*
 * Early initialization of comm channels. Must not rely
 * on interrupts, yet. Interrupt operation can be enabled
 * in _bsp_board_init().
 */
void
_bsp_init_ixp1200_comm(void)
{
    _bsp_queue_init(&uart_queue[0], uart_buffer[0], sizeof(uart_buffer[0]));
#ifdef UART_SPEED_AT_9600
    init_uart((void *) 0, 9600);
#else //#ifdef UART_SPEED_AT_9600
    init_uart((void *) 0, 38400);
#endif //#ifdef UART_SPEED_AT_9600

#ifdef DO_LED
    write_led (0x19);
#endif // DO_LED
}

#if DEBUG_UART_FUNCTIONS
/*
 * Print out a single Hex digit
 */
void uart_putnibble(void *base, unsigned char printval)
{
    if (printval <= 9)
        uart_putchar(base, printval + '0');
    else if ((printval >= 0xA) && (printval <= 0xF))
        uart_putchar(base, printval - 0xA + 'A');
    else
        uart_putchar(base, '.');
}

/*
 * Print out a byte value
 */
void uart_putbyte(void *base, unsigned char printval)
{
    uart_putnibble(base, (unsigned char)((printval >> 4)  & 0x0000000F));
    uart_putnibble(base, (unsigned char)((printval >> 0)  & 0x0000000F));
}

/*
 * Print out a word value
 */
void uart_putword(void *base, unsigned short printval)
{
    uart_putnibble(base, (unsigned char)((printval >> 12) & 0x0000000F));
    uart_putnibble(base, (unsigned char)((printval >> 8)  & 0x0000000F));
    uart_putnibble(base, (unsigned char)((printval >> 4)  & 0x0000000F));
    uart_putnibble(base, (unsigned char)((printval >> 0)  & 0x0000000F));
}

/*
 * Print out a long value
 */
void uart_putlong(void *base, unsigned long printval)
{
    uart_putnibble(base, (unsigned char)((printval >> 28) & 0x0000000F));
    uart_putnibble(base, (unsigned char)((printval >> 24) & 0x0000000F));
    uart_putnibble(base, (unsigned char)((printval >> 20) & 0x0000000F));
    uart_putnibble(base, (unsigned char)((printval >> 16) & 0x0000000F));
    uart_putnibble(base, (unsigned char)((printval >> 12) & 0x0000000F));
    uart_putnibble(base, (unsigned char)((printval >> 8)  & 0x0000000F));
    uart_putnibble(base, (unsigned char)((printval >> 4)  & 0x0000000F));
    uart_putnibble(base, (unsigned char)((printval >> 0)  & 0x0000000F));
}


/*
 * Print out a NULL terminated string
 */
void uart_putstring(void *base, char *string)
{
    if (string != NULL)
    {
        while (*string != '\0')
        {
            uart_putchar(base, *string);
            string++;
        }
    }
}

/*
 * Print out a single character
 */
void uart_putcharacter(void *base, char c)
{
    uart_putchar(base, c);
}
#endif /* DEBUG_UART_FUNCTIONS */


/*
 *  Flush Icache. Entire cache.
 */
void __icache_flush(void *__p, int __nbytes)
{
    __mcr(ARM_CACHE_COPROCESSOR_NUM,
          ARM_COPROCESSOR_OPCODE_DONT_CARE,
          0,
          ARM_CACHE_OPERATIONS_REGISTER,
          ARM_FLUSH_CACHE_INST_RM,
          ARM_FLUSH_CACHE_INST_OPCODE);
}

/*
 *  Flush entire Dcache.
 */
void __dcache_flush(void *__p, int __nbytes)
{
    register volatile unsigned long *ptr = REG32_PTR(IXP1200_ZEROS_BANK_BASE);
    register volatile unsigned long *ptr_stop =
         REG32_PTR(IXP1200_ZEROS_BANK_BASE + IXP1200_DCACHE_SIZE);
    register unsigned long temp;

    /*
     * Do a load to each line of the cache to
     * force a write-back
     */
    while (ptr < ptr_stop)
    {
	temp = *ptr;
	ptr += IXP1200_DCACHE_LINESIZE_WORDS;
    }

    /*
     * Now invalidate the cache
     */
    __mcr(ARM_CACHE_COPROCESSOR_NUM,
	  ARM_COPROCESSOR_OPCODE_DONT_CARE,
	  temp,
	  ARM_CACHE_OPERATIONS_REGISTER,
	  ARM_FLUSH_CACHE_DATA_RM,
	  ARM_FLUSH_CACHE_DATA_OPCODE);
}

