/*
 * xpflash.c -- Support for Intel(R) IXP1200 Evaluation Board
 *
 * Copyright (c) 1998, 1999 Cygnus Support
 *
 * The authors hereby grant permission to use, copy, modify, distribute,
 * and license this software and its documentation for any purpose, provided
 * that existing copyright notices are retained in all copies and that this
 * notice is included verbatim in any distributions. No written agreement,
 * license, or royalty fee is required for any of the authorized uses.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * Intel is a Registered Trademark of Intel Corporation.
 * ARM is a Registered Trademark of Advanced RISC Machines Limited.
 * Other Brands and Trademarks are the property of their respective owners.
 */

#include __BOARD_HEADER__
#include <string.h>
#include <bsp/bsp.h>
#include <bsp/cpu.h>
#include "net.h"
#include "bsp_if.h"
#include <bsp/platform.h>
#include <bsp/ixpflash.h>

static unsigned int flash_base = 0;  //mmu off, flash ROM starting address is 0
//static unsigned int flash_size = 2 * 1024 * 1024;    //2MB

static int dev_write(int offset, const void *buf, int size);
static int  dev_erase(int offset);
static bool dev_identify(void);

static void f28_write(volatile unsigned int *lptr, unsigned int data);
static void f28_erase(volatile unsigned int *lptr);
static void f28_reset(void);
static unsigned int f28_status(void);
static void f28_clear_status(void);
static unsigned int f28_read_id(volatile unsigned int *lptr);

extern void *malloc(int size);
extern void free(void *buff);
extern void change_pc_to_tmp_store();  

/* debug print options */
//#define IXPFLAH_DEBUG_INFO      1


unsigned int flash_valid_check=0;
unsigned int flash_is_valid=0;
unsigned int flash_rom_type_supported=0;
int ip_assign_method=0;
unsigned int lip_bootp=0;
unsigned int rip_bootp=0;
unsigned int lip_flash=0;
unsigned int rip_flash=0;
char kernel_file_flash[80];
char ramdisk_file_flash[80];
unsigned int kernel_load_offset_flash=0;
char cmd_line_flash[80];
int countdown_flash=0; 

void get_parameters_in_flash();
int save_parameters_into_flash();
void test_flash_rom_identify(void);
void get_boot_option_from_flash();

int bo_assign_method=0;
unsigned int bo_lip=0;
unsigned int bo_rip=0;
char bo_kernel_file[80]="";
char bo_ramdisk_file[80]="";
unsigned int bo_kernel_load_offset=0;
char bo_cmd_line[80]="";
unsigned int bo_countdown=0; 
int ixp1200_uart_poll_read(char *ch);
void set_ixp1200_pc(unsigned int desired_pc);
unsigned int minimum_linux_kernel_load_offset=MINIMUM_LINUX_KERNEL_LOAD_OFFSET; 
void start_linux_booting();
 
union uint_fname
{
    unsigned int u_int[20];
    char fname[80];
};

void test_flash_rom_identify(void)
{
    dev_identify();
}

void get_parameters_in_flash()
{
 int i=0;
 union uint_fname fn_tmp;
  
  flash_is_valid=0;	//default
  flash_rom_type_supported=0;  //default
  if(dev_identify())
      flash_rom_type_supported=1;
  ip_assign_method=ASSIGNED_MANUALLY;	//default  
  lip_bootp=0;
  rip_bootp=0;
  lip_flash=0;
  rip_flash=0;
  kernel_file_flash[0]=0;
  ramdisk_file_flash[0]=0;
  kernel_load_offset_flash=0;
  cmd_line_flash[0]=0;
  if(flash_rom_type_supported)
  {
    flash_valid_check=*FLASH_VALID_CHECK;
#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf(" flash_valid_check=%8.8X\n", flash_valid_check);
#endif //#ifdef IXPFLAH_DEBUG_INFO
    if(flash_valid_check == FLASH_VALID)
    {
        flash_is_valid=1;
        ip_assign_method=*IP_ASSIGN_METHOD;
#ifdef IXPFLAH_DEBUG_INFO
        bsp_printf(" ip_assign_method=%8.8X\n", ip_assign_method);
#endif //#ifdef IXPFLAH_DEBUG_INFO
        lip_flash=*LOCAL_MACHINE_IP;
        rip_flash=*REMOTE_SERVER_IP;

        //get kernel_file_flash (occupies 80 bytes from LINUX_KERNEL_FILE
        for(i=0; i< 20; i++)
        {
#ifdef __LITTLE_ENDIAN__       
            fn_tmp.u_int[i]= *(unsigned int *)(LINUX_KERNEL_FILE + i);
#else //#ifdef __LITTLE_ENDIAN__       
            fn_tmp.u_int[i]= swap32(*(unsigned int *)(LINUX_KERNEL_FILE + i));
#endif //#ifdef __LITTLE_ENDIAN__  
       }
        strcpy(kernel_file_flash, fn_tmp.fname);      
#ifdef IXPFLAH_DEBUG_INFO
        bsp_printf("Get fn:%s, u[0]=%8.8X, u[1]=%8.8X, u[2]=%8.8X\n",
          fn_tmp.fname, fn_tmp.u_int[0], fn_tmp.u_int[1], fn_tmp.u_int[2]);
#endif //#ifdef IXPFLAH_DEBUG_INFO

        //get ramdisk_file_flash (occupies 80 bytes from LINUX_RAMDISK_FILE
        for(i=0; i< 20; i++)
        {
#ifdef __LITTLE_ENDIAN__       
            fn_tmp.u_int[i]= *(unsigned int *)(LINUX_RAMDISK_FILE + i);
#else //#ifdef __LITTLE_ENDIAN__       
            fn_tmp.u_int[i]= swap32(*(unsigned int *)(LINUX_RAMDISK_FILE + i));
#endif //#ifdef __LITTLE_ENDIAN__       
        }
        strcpy(ramdisk_file_flash, fn_tmp.fname);      
#ifdef IXPFLAH_DEBUG_INFO
        bsp_printf("Get fn:%s, u[0]=%8.8X, u[1]=%8.8X, u[2]=%8.8X\n",
          fn_tmp.fname, fn_tmp.u_int[0], fn_tmp.u_int[1], fn_tmp.u_int[2]);
#endif //#ifdef IXPFLAH_DEBUG_INFO

        //get kernel_load_offset_flash
        kernel_load_offset_flash=*LINUX_KERNEL_LOAD_OFFSET;

        //get cmd_line_flash (occupies 80 bytes from LINUX_CMD_LINE
        for(i=0; i< 20; i++)
        {
#ifdef __LITTLE_ENDIAN__       
            fn_tmp.u_int[i]= *(unsigned int *)(LINUX_CMD_LINE + i);
#else //#ifdef __LITTLE_ENDIAN__       
            fn_tmp.u_int[i]= swap32(*(unsigned int *)(LINUX_CMD_LINE + i));
#endif //#ifdef __LITTLE_ENDIAN__       
        }
        strcpy(cmd_line_flash, fn_tmp.fname);      
#ifdef IXPFLAH_DEBUG_INFO
        bsp_printf("Get fn:%s, u[0]=%8.8X, u[1]=%8.8X, u[2]=%8.8X\n",
          fn_tmp.fname, fn_tmp.u_int[0], fn_tmp.u_int[1], fn_tmp.u_int[2]);
#endif //#ifdef IXPFLAH_DEBUG_INFO

        //get countdown_flash
        countdown_flash=*LINUX_BOOT_COUNTDOWN;
    }
  }
  //setup initial boot options 
  get_boot_option_from_flash();
}

void get_boot_option_from_flash()
{
  bo_assign_method=ip_assign_method;
  if(bo_assign_method == ASSIGNED_MANUALLY)
  {  
      bo_lip=lip_flash;
      bo_rip=rip_flash;
  }
  else 
  {
      bo_lip=lip_bootp;
      bo_rip=rip_bootp;
  }
  strcpy(bo_kernel_file, kernel_file_flash);
  if(bo_kernel_file[0] == 0)
      strcpy(bo_kernel_file, "zImage");
  strcpy(bo_ramdisk_file, ramdisk_file_flash);
  if(bo_ramdisk_file[0] == 0)
      strcpy(bo_ramdisk_file, "ramdisk_img.gz");
  bo_kernel_load_offset=kernel_load_offset_flash;
  if(bo_kernel_load_offset < minimum_linux_kernel_load_offset)
      bo_kernel_load_offset=DEFAULT_LINUXKERNELLOADADDRESS;
     
  strcpy(bo_cmd_line, cmd_line_flash);
  if(bo_cmd_line[0] == 0)
      strcpy(bo_cmd_line, "");	//defaut is NULL now
  bo_countdown=countdown_flash;
  if(bo_countdown < DEFAULT_LINUX_BOOT_COUNTDOWN) 
      bo_countdown=DEFAULT_LINUX_BOOT_COUNTDOWN; 
}

int save_parameters_into_flash()
{
    unsigned int src_addr=IXPFLASH_PARAM_BLOCK_START_ADDR;
    unsigned int *src_buf=NULL;
    int i=0;
    union uint_fname fn_tmp;

    flash_valid_check=FLASH_VALID;
    if( (flash_valid_check==*FLASH_VALID_CHECK) && 
        (ip_assign_method==bo_assign_method) && 
        (lip_flash==bo_lip) && 
        (rip_flash==bo_rip) &&
        (strcmp(kernel_file_flash, bo_kernel_file)==0) &&
        (strcmp(ramdisk_file_flash, bo_ramdisk_file)==0) &&
        (kernel_load_offset_flash==bo_kernel_load_offset) &&
        (strcmp(cmd_line_flash, bo_cmd_line)==0) &&
        (countdown_flash==bo_countdown) )
    {
#ifdef IXPFLAH_DEBUG_INFO
        bsp_printf("No change, no need to save!\n");
#endif //#ifdef IXPFLAH_DEBUG_INFO
        flash_is_valid=1;
        return 0;   //success    
    }

    //read back all stuff on the block, 
    src_buf=(unsigned int *)(malloc(IXPFLASH_PARAM_BLOCK_SIZE));
    if(src_buf==NULL) return 1;  //fail

    ip_assign_method=bo_assign_method;
                
    if(!ip_assign_method)
    {
        lip_flash=bo_lip;
        rip_flash=bo_rip;
    }
    else 
    {
        lip_flash=0;
        rip_flash=0;
    }
    strcpy(kernel_file_flash, bo_kernel_file);
    strcpy(ramdisk_file_flash, bo_ramdisk_file);
    kernel_load_offset_flash=bo_kernel_load_offset;
    strcpy(cmd_line_flash, bo_cmd_line);

    countdown_flash=bo_countdown;

    flash_is_valid=1;
    for(i=0; i < (IXPFLASH_PARAM_BLOCK_SIZE/4); i++)
    {
        src_buf[i]= *(unsigned int *)src_addr;
        src_addr +=4;
    }

#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf("Before change: src[%d]=%8.8X, src[%d]=%8.8X, src[%d]=%8.8X\n",
        ((IXPFLASH_CYGMON_PARAM_OFFSET + FLASH_VALID_CHECK_o)/4),
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + FLASH_VALID_CHECK_o)/4)],
        ((IXPFLASH_CYGMON_PARAM_OFFSET + IP_ASSIGN_METHOD_o)/4),
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + IP_ASSIGN_METHOD_o)/4)],
        ((IXPFLASH_CYGMON_PARAM_OFFSET + LOCAL_MACHINE_IP_o)/4),
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + LOCAL_MACHINE_IP_o)/4)]);
    bsp_printf("Before change: src[%d]=%8.8X, src[%d]=%8.8X, src[%d]=%8.8X\n",
        ((IXPFLASH_CYGMON_PARAM_OFFSET + REMOTE_SERVER_IP_o +4)/4),
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + REMOTE_SERVER_IP_o +4)/4)],
        ((IXPFLASH_CYGMON_PARAM_OFFSET + REMOTE_SERVER_IP_o + 8)/4),
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + REMOTE_SERVER_IP_o + 8)/4)],
        ((IXPFLASH_PARAM_BLOCK_SIZE/4) -1),  src_buf[((IXPFLASH_PARAM_BLOCK_SIZE/4)-1)]);
#endif //#ifdef IXPFLAH_DEBUG_INFO
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + FLASH_VALID_CHECK_o)/4)]=flash_valid_check;
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + IP_ASSIGN_METHOD_o)/4)]=bo_assign_method;
        if(!bo_assign_method)
        {
            src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + LOCAL_MACHINE_IP_o)/4)]=bo_lip;
            src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + REMOTE_SERVER_IP_o)/4)]=bo_rip;
        }
        else
        {
            src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + LOCAL_MACHINE_IP_o)/4)]=0;
            src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + REMOTE_SERVER_IP_o)/4)]=0;
        }

        //handle kernel file name
        for(i=0; i< 20; i++)   
            fn_tmp.u_int[i]= 0;	//reset
        strcpy(fn_tmp.fname, bo_kernel_file);
#ifdef IXPFLAH_DEBUG_INFO
        bsp_printf("fn:%s, u[0]=%8.8X, u[1]=%8.8X\n", fn_tmp.fname, fn_tmp.u_int[0], fn_tmp.u_int[1]);
#endif //#ifdef IXPFLAH_DEBUG_INFO
        for(i=0; i< 20; i++)
        {

#ifndef __LITTLE_ENDIAN__       
            fn_tmp.u_int[i]= swap32(fn_tmp.u_int[i]);
#endif //#ifndef __LITTLE_ENDIAN__  

            src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + LINUX_KERNEL_FILE_o)/4)+i]=
                fn_tmp.u_int[i];
        }

        //handle ramdisk file name
        for(i=0; i< 20; i++)   
            fn_tmp.u_int[i]= 0;	//reset
        strcpy(fn_tmp.fname, bo_ramdisk_file);
#ifdef IXPFLAH_DEBUG_INFO
        bsp_printf("fn:%s, u[0]=%8.8X, u[1]=%8.8X\n", fn_tmp.fname, fn_tmp.u_int[0], fn_tmp.u_int[1]);
#endif //#ifdef IXPFLAH_DEBUG_INFO
        for(i=0; i< 20; i++)
        {

#ifndef __LITTLE_ENDIAN__       
            fn_tmp.u_int[i]= swap32(fn_tmp.u_int[i]);
#endif //#ifndef __LITTLE_ENDIAN__  

            src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + LINUX_RAMDISK_FILE_o)/4)+i]=
                fn_tmp.u_int[i];
        }

        //handle kernel_load_offset
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + LINUX_KERNEL_LOAD_OFFSET_o)/4)]=bo_kernel_load_offset;

        //handle command line string
        for(i=0; i< 20; i++)   
            fn_tmp.u_int[i]= 0;	//reset
        strcpy(fn_tmp.fname, bo_cmd_line);
#ifdef IXPFLAH_DEBUG_INFO
        bsp_printf("fn:%s, u[0]=%8.8X, u[1]=%8.8X\n", fn_tmp.fname, fn_tmp.u_int[0], fn_tmp.u_int[1]);
#endif //#ifdef IXPFLAH_DEBUG_INFO
        for(i=0; i< 20; i++)
        {

#ifndef __LITTLE_ENDIAN__       
            fn_tmp.u_int[i]= swap32(fn_tmp.u_int[i]);
#endif //#ifndef __LITTLE_ENDIAN__  

            src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + LINUX_CMD_LINE_o)/4)+i]=
                fn_tmp.u_int[i];
        }
 
        //handle linux kernel boot countdown 
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + LINUX_BOOT_COUNTDOWN_o)/4)]=bo_countdown;

#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf("After change: src[%d]=%8.8X, src[%d]=%8.8X, src[%d]=%8.8X\n",
        ((IXPFLASH_CYGMON_PARAM_OFFSET + FLASH_VALID_CHECK_o)/4),
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + FLASH_VALID_CHECK_o)/4)],
        ((IXPFLASH_CYGMON_PARAM_OFFSET + IP_ASSIGN_METHOD_o)/4),
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + IP_ASSIGN_METHOD_o)/4)],
        ((IXPFLASH_CYGMON_PARAM_OFFSET + LOCAL_MACHINE_IP_o)/4),
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + LOCAL_MACHINE_IP_o)/4)]);
    bsp_printf("After change: src[%d]=%8.8X, src[%d]=%8.8X, src[%d]=%8.8X\n",
        ((IXPFLASH_CYGMON_PARAM_OFFSET + REMOTE_SERVER_IP_o +4)/4),
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + REMOTE_SERVER_IP_o + 4)/4)],
        ((IXPFLASH_CYGMON_PARAM_OFFSET + REMOTE_SERVER_IP_o + 8)/4),
        src_buf[((IXPFLASH_CYGMON_PARAM_OFFSET + REMOTE_SERVER_IP_o + 8)/4)],
        ((IXPFLASH_PARAM_BLOCK_SIZE/4) -1),  src_buf[((IXPFLASH_PARAM_BLOCK_SIZE/4)-1)]);
#endif //#ifdef IXPFLAH_DEBUG_INFO

    dev_erase(IXPFLASH_PARAM_BLOCK_START_ADDR);
#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf("Finished dev_erase.\n");
#endif //#ifdef IXPFLAH_DEBUG_INFO
    dev_write(IXPFLASH_PARAM_BLOCK_START_ADDR, src_buf, (IXPFLASH_PARAM_BLOCK_SIZE/4));
#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf("Finished dev_write.\n");
#endif //#ifdef IXPFLAH_DEBUG_INFO

    free(src_buf);
}

void start_linux_booting()
{
 
    *(volatile unsigned int *)TMP_STORE_AT_SDRAM_FOR_BOOT_LINUX=bo_kernel_load_offset;
    change_pc_to_tmp_store();  
}

/*****************************************************************************

Device Flash: Actual low level functions. These are the one only functions
that directly deal with the Flash device. These set of functions have
knowledge about

1. The device organisation. The no. of 8K blocks, 64K blocks etc.
2. interleaving of two 16 bit flash devices to get a 32 bit flash 
3. The size of the flash (2MB = 1MB+1MB)
4. The Base address at which the flash is present(FLASH_BASE 0xF3000000)
5. The specific algorithms to identify, write & erase the device.
 
These functions are named dev_*. All the dev_* functions operate in the 
address (offset) range of 0 to flash_size. All these 
offsets/addresses must be 32bit aligned. It is the responsibility of the 
caller to ensure this.
*****************************************************************************/



/*
 Limitations: offset must be block aligned. 
 buf must be 32 bit aligned. 
 size must be a multiple of 4. 
 */
static int dev_write(int offset, const void *buf, int size)
{
  unsigned int status;
  unsigned int *lptr = (unsigned int *)(flash_base + offset);
  unsigned int *lbuf = (unsigned int *)buf; /* dangerous!?!?!? To be guaranteed by caller*/
#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf("dev_write size=%d\n", size);
#endif //#ifdef IXPFLAH_DEBUG_INFO

  while(size--) 
    f28_write(lptr++ ,*lbuf++);

#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf("dev_write - finish f28_write\n");
#endif //#ifdef IXPFLAH_DEBUG_INFO
  /* check for errors */
  status = f28_status();
  
  /* error in first device || error in second device  */
  if((status & (BIT(5) | BIT(4) | BIT(3) | BIT(1))) ||
     ((status>>16) & (BIT(5) | BIT(4) | BIT(3) | BIT(1)))) {
    bsp_printf("dev_write: write error at offset %lx, size %lx : status %lx\n", offset, size,status);
    size =0; /* retval */
    f28_clear_status();
  }

  /* reset to read array mode */
  f28_reset();

  return size;
}

static int dev_erase(int offset)
{
  int ret_val = 0;

  unsigned int status;
  unsigned int *lptr = (unsigned int *)(flash_base+offset);
#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf("dev_erase - offset=%lx\n", offset);
#endif //#ifdef IXPFLAH_DEBUG_INFO

#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf("dev_erase - erasing block %lx\n", (unsigned int)lptr);
#endif //#ifdef IXPFLAH_DEBUG_INFO
    f28_erase(lptr);
#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf("dev_erase - finished f28_erase\n");
#endif //#ifdef IXPFLAH_DEBUG_INFO

  /* check for errors */
  status = f28_status();
  
  /* error in first device || error in second device  */
  if((status & (BIT(5) | BIT(4) | BIT(3) | BIT(1))) ||
     ((status>>16) & (BIT(5) | BIT(4) | BIT(3) | BIT(1)))) {
    bsp_printf("dev_erase: erase error at offset %lx : status %lx\n",offset,status);
    ret_val = -1; /* define error code(s) for this function */ 
    f28_clear_status();
  }

#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf("dev_erase - finished f28_clear_status\n");
#endif //#ifdef IXPFLAH_DEBUG_INFO
  /* reset to read array mode */
  f28_reset();

  return ret_val;
}

static bool dev_identify(void)
{
    bool retval = true;
    unsigned int *lptr = (unsigned int *)flash_base;
    unsigned int l_mfrid = f28_read_id(lptr);
    unsigned int l_devid = f28_read_id(lptr+1);
    unsigned short mfrid = (unsigned short)l_mfrid;
    unsigned short devid = (unsigned short)l_devid;
  
//    bsp_printf("l_mfrid=%8.8X, l_devid=%8.8X\n", l_mfrid, l_devid);

    if(mfrid != F28_MANUF_ID || devid != FTYPE_28F800B)
    {
//        bsp_printf(" Unable to detect flash device\n");
        retval = false;
    }

    f28_reset();
  
    return retval;
}

 


/* Try to make all the f28_* functions inline */
static void f28_write(volatile unsigned int *lptr, unsigned int data)
{
  *lptr = F28CMD_PROGRAM_SETUP;
  *lptr = data;

  /* after writing the flash will always be in read status register mode */

  while(!(*lptr & BIT(7))) /* status of the first device */
	;

  while (!((*lptr >> 16) & BIT(7))) /* status of the second device */
    ;

} 

static void f28_erase(volatile unsigned int *lptr)
{
  *lptr = F28CMD_ERASE_SETUP;
  *lptr = F28CMD_ERASE_CONFIRM;

  /* after erasing, the flash will always be in read status register mode */

  while(!(*lptr & BIT(7))) /* status of the first device */
#ifndef IXPFLAH_DEBUG_INFO
    ;
#else //#ifndef IXPFLAH_DEBUG_INFO
  {
    bsp_printf("*%lx=%lx\n", (unsigned int)lptr, *lptr);
  }
#endif //#ifndef IXPFLAH_DEBUG_INFO

  while (!((*lptr >> 16) & BIT(7))) /* status of the second device */
#ifndef IXPFLAH_DEBUG_INFO
    ;
#else //#ifndef IXPFLAH_DEBUG_INFO
  {
    bsp_printf("*%lx=%lx\n", (unsigned int)lptr, *lptr);
  }
#endif //#ifndef IXPFLAH_DEBUG_INFO
}

static void f28_reset(void)
{
  volatile unsigned int *lptr = (unsigned int *)flash_base;

  *lptr = F28CMD_RESET;
}

static unsigned int f28_status(void)
{
  volatile  unsigned int *lptr = (unsigned int *)flash_base;

  *lptr = F28CMD_READ_STATUS;

  return *lptr;
}

static void f28_clear_status(void)
{
  volatile unsigned int *lptr = (unsigned int *)flash_base;
 
  *lptr = F28CMD_CLEAR_STATUS;
}

static unsigned int f28_read_id(volatile unsigned int *lptr)
{
#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf(" BOR of f28_read_id: lptr=%8.8X\n", (unsigned int)lptr);
#endif //#ifndef IXPFLAH_DEBUG_INFO
    *lptr = F28CMD_READ_ID;
  
#ifdef IXPFLAH_DEBUG_INFO
    bsp_printf(" f28_read_id - finish F28CMD_READ_ID: *lptr=%8.8X\n", *lptr);
#endif //#ifndef IXPFLAH_DEBUG_INFO
    return *lptr;

}

int ixp1200_uart_poll_read(char *ch)
{
    //unsigned int status=*(volatile unsigned int *)UART_SR;
    unsigned int data;

    //even uart not RX_READY, we can still get the input character
    //bsp_printf("st=%8.8X ",status);
    //if (!(status & UART_RX_READY))
    //    return -1;


    data=*(volatile unsigned int *)UART_DR;
    *ch=(char)(data & 0x000000FF);
    // bsp_printf("data=%8.8X, ch=%2.2X;", data, *ch);
    return 0;
}
  
