/*
 * pci_cfg.c -- Support for Intel(R) IXP1200 Evaluation Board
 *
 * Copyright (c) 1998, 1999 Cygnus Support
 *
 * The authors hereby grant permission to use, copy, modify, distribute,
 * and license this software and its documentation for any purpose, provided
 * that existing copyright notices are retained in all copies and that this
 * notice is included verbatim in any distributions. No written agreement,
 * license, or royalty fee is required for any of the authorized uses.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * Intel is a Registered Trademark of Intel Corporation.
 * ARM is a Registered Trademark of Advanced RISC Machines Limited.
 * Other Brands and Trademarks are the property of their respective owners.
 */

#include __BOARD_HEADER__
#include <string.h>
#include <bsp/bsp.h>
#include <bsp/cpu.h>
#include "bsp_if.h"
#include <bsp/platform.h>
#include <bsp/pci_cfg.h>

/* debug print options */
//#define DEBUG_PRINT_PCI_CONFIG_INFO 	1 

//#define OUTPUT_SA_CONTROL_DEBUG_INFO 1 

//#define SHOW_VID_DID_REV	1 

/* set PCI_MAX_QUERY_DEV to 9, due to IXP1200 hang in querying device 9 */
#define PCI_MAX_QUERY_DEV 	9

#define MAX_PCI_DEVICES	10

#define MEM_MAP		0
#define IO_MAP 		1

#define PCI_IN_UINT(x)      *(unsigned int *)(x)
#define PCI_OUT_UINT(x,y)   *(unsigned int *)(x) = (unsigned int) y

int is_ixp1200_central_func=0; /* global variable */
unsigned int ixp1200_chip_rev=0; 

static int total_devs=0;
static int pci_conf_addr_base = IXP1200_PCI_TYPE_0_CONFIG_BASE;
static unsigned int next_avl_io_addr=0;   
static unsigned int next_avl_mem_addr=0;
static struct pci_dev pci_devs[MAX_PCI_DEVICES];

static unsigned int get_pci_io_addr(struct pci_dev *pdev, int where);
static void delay_few_usecs(void);
static void assign_pdev_addr(struct pci_dev *pdev, int offset, int type, 
	unsigned int size);

extern void write_led(char ch);
extern void ixp1200_sw_delay_msec(int msec);

/******************************************************************************
* pci_read_config_byte - read one byte from pci configuration space
*/
int pci_read_config_byte(struct pci_dev *pdev, int where, unsigned char *pval)
{
    unsigned int uint_val=0;
    int byte_pos=0;

    /* need to get dword first */
    pci_read_config_dword(pdev, (where & 0xFFFFFFFC), &uint_val);
    byte_pos = where % 4;

    *pval = (unsigned char)( (uint_val >> (8*byte_pos)) & 0xFF);

    return OK;
}

/******************************************************************************
* pci_read_config_word - read one word from pci configuration space
*/
int pci_read_config_word(struct pci_dev *pdev, int where, unsigned short *pval)
{
    unsigned int uint_val=0;
    int word_pos=0;

    /* need to get dword first */
    pci_read_config_dword(pdev, (where & 0xFFFFFFFC), &uint_val);
    word_pos = where % 4;

    *pval = (unsigned short)( (uint_val >> (8*word_pos)) & 0xFFFF);

    return OK;
}


/******************************************************************************
* pci_read_config_dwod - read one doule word from pci configuration space
*/
int pci_read_config_dword(struct pci_dev *pdev, int where, unsigned int *pval)
{
    *pval= (unsigned int)(PCI_IN_UINT(get_pci_io_addr(pdev, where)));
    return OK;
}

/******************************************************************************
* pci_write_config_byte - write one byte to pci configuration space
*/
int pci_write_config_byte(struct pci_dev *pdev, int where, unsigned char val)
{
    unsigned int uint_val=0, addr=0;
    int byte_pos=0;

    byte_pos = where % 4;
    addr =  get_pci_io_addr(pdev, (where & 0xFFFFFFFC));
    uint_val = (unsigned int)(PCI_IN_UINT(addr));
    uint_val &= ~(0xFF << (byte_pos*8));  /* mask out */
    uint_val |= ( ((unsigned int)val & 0xFF) << (byte_pos*8) );
    PCI_OUT_UINT(addr, uint_val);
    return OK;
}

/******************************************************************************
* pci_write_config_word - write one word to pci configuration space
*/
int pci_write_config_word(struct pci_dev *pdev, int where, unsigned short val)
{
    unsigned int uint_val=0, addr=0;
    int word_pos=0;

    word_pos = where % 4;
    addr =  get_pci_io_addr(pdev, (where & 0xFFFFFFFC));
    uint_val = (unsigned int)(PCI_IN_UINT(addr));
    uint_val &= ~(0xFFFF << (word_pos*8));  /* mask out */
    uint_val |= ( ((unsigned int)val & 0xFFFF) << (word_pos*8) );
    PCI_OUT_UINT(addr, uint_val);
    return OK;
}


/******************************************************************************
* pci_write_config_dword- write one double word to pci configuration space
*/
int pci_write_config_dword(struct pci_dev *pdev, int where, unsigned int val)
{
    unsigned int addr=0;

    addr =  get_pci_io_addr(pdev, where);
    PCI_OUT_UINT(addr, val);
    return OK;
}

/******************************************************************************
*  get_pci_io_addr - return an address used for pci configuration space access)
*/
static unsigned int get_pci_io_addr(struct pci_dev *pdev, int where)
{
    /* this only work for bus 0*/
    unsigned int addr=0;

    addr  = (unsigned int)pci_conf_addr_base;	 /* get the base */
    addr |= (unsigned int)(1 << ( ((pdev->devfn >> 3) & 0x1F) + 16) );
    addr |= (unsigned int)((pdev->devfn & 0x7) << 8);
    addr |= (unsigned int)(where & 0xFC);
    return addr;
}


/******************************************************************************
* pci_find_device - return a pointer to sruct pi_devs[i] which have the info
*                   requested
*/
struct pci_dev *pci_find_device(int vendor_id, int device_id, int index)
{
    /* since we already have all devices info in pci_devs[], we can just
       find device from there, not need to emurate the whole pci bus again.
       input index starting from 0, means first occurrence of the device with
       the specified vendor_id and device_id
    */
    int i=0, occurrence=0;
    unsigned short s_vid=0, s_did=0;
    
    s_vid=(unsigned short)vendor_id;
    s_did=(unsigned short)device_id;

    for (i=0; i < total_devs; i++)
    {
	if( (pci_devs[i].vendor_id == s_vid) && (pci_devs[i].device_id == s_did) ) 
        {
	    if(index == occurrence) 
	    {
#ifdef DEBUG_PRINT_PCI_CONFIG_INFO
                bsp_printf(" device with vid=%04X, did=%04X index=%d found in 
                    pci_devs[%d] with bus=%d and dev=%d\n", s_vid, s_did, index,
                    i, pci_devs[i].bus_no, ((pci_devs[i].devfn >> 3) & 0x1F) ); 
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */
		return &pci_devs[i];
	    }		
            occurrence++;
         } 
    }
    
#ifdef DEBUG_PRINT_PCI_CONFIG_INFO 
    bsp_printf(" device with vid=%04X, did=%04X index=%d not found.\n", s_vid, 
	s_did, index);
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */
    return NULL;
}

/******************************************************************************
* ixp1200_pci_config - detect PCI devices on backplane and assign the resources
* they requested
*/
void ixp1200_pci_config(void)
{
    struct pci_dev cur_dev;
    unsigned short us_val;
    unsigned int uint_val=0, req_size=0;
    int dev, i, type;

#ifdef DEBUG_PRINT_PCI_CONFIG_INFO
    bsp_printf("In ixp1200_pci_config(), it is cental_func.\n");
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */

    next_avl_mem_addr = 0;	/* start with zero */
    next_avl_io_addr = 0;	/* start with zero */    

#ifdef DEBUG_PRINT_PCI_CONFIG_INFO
    /* before assign pci resources, display first */
    bsp_printf(" before assign pci resources, display first: \n");
    bsp_printf("  BAR_0=%08X, BAR_1=%08X, BAR_2=%08X\n", *IXP1200_PCI_MEM_BAR, 
	*IXP1200_PCI_IO_BAR, *IXP1200_PCI_DRAM_BAR);
    bsp_printf("  BAR_3=%08X, BAR_4=%08X, BAR_5=%08X\n", *IXP1200_PCI_BAR_RSV_1, 
	*IXP1200_PCI_BAR_RSV_2, *IXP1200_PCI_BAR_RSV_3);
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */

    /* assign pci resources to ixp1200 - memory and i/o space */
    /* assign SDRAM start address as 0 */
    *IXP1200_PCI_DRAM_BAR=next_avl_mem_addr | (*IXP1200_PCI_DRAM_BAR & 0x0F);
    next_avl_mem_addr += IXP1200EB_SDRAM_SIZE; 


    /* assign CSR memory address to next available memory address */
    *IXP1200_PCI_MEM_BAR=next_avl_mem_addr | (*IXP1200_PCI_MEM_BAR & 0x0F); 
    next_avl_mem_addr += IXP1200EB_CSR_MEM_SIZE;
    
    /* assign CSR I/O address to 0 */
    *IXP1200_PCI_IO_BAR=next_avl_io_addr | (*IXP1200_PCI_IO_BAR & 0x07);
    next_avl_io_addr += IXP1200EB_CSR_IO_SIZE;

    *IXP1200_PCI_CMD_STAT= (*IXP1200_PCI_CMD_STAT & 0xFFFF0000) |
        PCI_COMMAND_IO | PCI_COMMAND_MEMORY | PCI_COMMAND_MASTER | 
        PCI_COMMAND_INVALIDATE;   

#ifdef DEBUG_PRINT_PCI_CONFIG_INFO
    /* after assign pci resources, display first */
    bsp_printf(" after assign pci resources, display first: \n");
    bsp_printf("  BAR_0=%08X, BAR_1=%08X, BAR_2=%08X\n", *IXP1200_PCI_MEM_BAR, 
	*IXP1200_PCI_IO_BAR, *IXP1200_PCI_DRAM_BAR);
    bsp_printf("  BAR_3=%08X, BAR_4=%08X, BAR_5=%08X\n", *IXP1200_PCI_BAR_RSV_1, 
	*IXP1200_PCI_BAR_RSV_2, *IXP1200_PCI_BAR_RSV_3);
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */

    total_devs = 0;

    /* detect pci devices on backplane and assign pci resources they
       requested, only check devices on backplane  and func 0 */
    /* set PCI_MAX_QUERY_DEV to 9, due to IXP1200 hang in querying device 9 */
    for (dev = 0; dev < PCI_MAX_QUERY_DEV; dev++ ) 
    {
       cur_dev.bus_no=0; 
       cur_dev.devfn=(int)((dev << 3) & 0xF8);
#ifdef DEBUG_PRINT_PCI_CONFIG_INFO 
	bsp_printf("pcicfg: querying dev%d", dev);
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */
        pci_read_config_word(&cur_dev, PCI_VENDOR_ID, &us_val);
#ifdef DEBUG_PRINT_PCI_CONFIG_INFO
        bsp_printf(" VID=%08X\n", us_val);
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */
        if (us_val == 0xFFFF)
            continue;
        if (total_devs == PCI_MAX_QUERY_DEV)
            break;
        pci_devs[total_devs].bus_no = 0;
        pci_devs[total_devs].devfn = cur_dev.devfn;
        pci_devs[total_devs].vendor_id = us_val;
        pci_read_config_word(&cur_dev, PCI_DEVICE_ID, &us_val);
#ifdef DEBUG_PRINT_PCI_CONFIG_INFO
        bsp_printf("deviceid=%08X\n", us_val);
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */
        pci_devs[total_devs].device_id = us_val;
        for (i=0; i<6; i++) 
        {
            /* query IO_MAP or MEM_MAP and size */
            pci_write_config_dword(&cur_dev, 
                PCI_BASE_ADDRESS_0 + (4*i), 0xFFFFFFFF);
            pci_read_config_dword(&cur_dev, PCI_BASE_ADDRESS_0 + (4*i),
                            &uint_val);
#ifdef DEBUG_PRINT_PCI_CONFIG_INFO
            bsp_printf("requested address_0=%08X\n", uint_val);
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */
            if (uint_val == 0)  break; /* no more */
            if (uint_val & 1)
            {
                /* IO_MAP space */
                /* tmp = uint_val & 0xFFFFFFFC;
                size = ~(tmp-1);*/
                req_size= ~((uint_val & 0xFFFFFFFC) - 1);

                type=IO_MAP;
            }
            else
            {
                /* MEM_MAP space */
                /* tmp = uint_val & 0xFFFFFFF0; 
                size = ~(tmp-1); */
                req_size = ~((uint_val & 0xFFFFFFF0) - 1); 
                /* need to adjust next_avl_mem_addr to be the same order as 
                  requested, otherwise will round to the same order, maybe
                  overlapped with previous assigned memory map address, this 
                  will leave some hole in memory map sddress, it should not 
                  matter, because we should have enough memory spcae in this 
                  IXP1200EB backplane case */

                if( ((next_avl_mem_addr/req_size) * req_size) != 
		    next_avl_mem_addr) 
                {
#ifdef DEBUG_PRINT_PCI_CONFIG_INFO
                    bsp_printf(" next_avl_mem_addr=%08X, requested size=%08X,  
                        need round up\n", next_avl_mem_addr, req_size);
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */
                    next_avl_mem_addr=((next_avl_mem_addr/req_size) + 1)
                        * req_size;
#ifdef DEBUG_PRINT_PCI_CONFIG_INFO
                    bsp_printf(" after round up, next_avl_mem_addr=%08X \n", 
                        next_avl_mem_addr);
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */
                } 
                type=MEM_MAP;  
            }
            /* assign the resource */
            assign_pdev_addr(&cur_dev, i, type, req_size);
#ifdef DEBUG_PRINT_PCI_CONFIG_INFO
            pci_read_config_dword(&cur_dev, PCI_BASE_ADDRESS_0 + (4*i),
                            &uint_val);
            bsp_printf(" type=%d, size=%08X, BAR=%08X\n", type, req_size, 
		uint_val);
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */
        } /* end for i */
        /* enabel IO for this device */
        pci_write_config_word(&cur_dev, PCI_COMMAND, PCI_COMMAND_IO);
        /* assign INT_VEC */
        pci_write_config_dword(&cur_dev, PCI_INTERRUPT_LINE, 
	    IXP1200_PCI_INT_VEC);

        total_devs++;
    } /* end for dev */


    /* if mmu on, add page table entries as necessary */
 
#ifdef DEBUG_PRINT_PCI_CONFIG_INFO
    for (i=0; i<total_devs; i++) 
    {
        pci_read_config_word(&pci_devs[i], PCI_COMMAND, &us_val);
        bsp_printf("bus %d  dev %d  id %04X%04X, pci_command=%4.4X\n",
            pci_devs[i].bus_no, (pci_devs[i].devfn >> 3), pci_devs[i].vendor_id,
            pci_devs[i].device_id, us_val);
    } /* end for i */
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */

#ifdef OUTPUT_SA_CONTROL_DEBUG_INFO
    bsp_printf("Exit ixp1200_pci_config(), SA_Control = %08X\n",
               *IXP1200_PCI_SA_CONTROL);
#endif /* #ifdef OUTPUT_SA_CONTROL_DEBUG_INFO */
#ifndef __LITTLE_ENDIAN__
/* DC for chip stepping upto rev A3, following BE_BEO, BE_BEI not working 
    if(((*IXP1200_PCI_REV_CLASS) & 0xFF) == 1)
    {
        bsp_printf("ixp1200 rev. b0 in big endian , enable BE_BEO & BE_BEI, bf SA_Control=%08X\n",
               *IXP1200_PCI_SA_CONTROL);
        *IXP1200_PCI_SA_CONTROL= *IXP1200_PCI_SA_CONTROL | 0x00050000;
        bsp_printf(" after  SA_Control = %08X\n",
               *IXP1200_PCI_SA_CONTROL);
    }
*/
#endif //#ifndef __LITTLE_ENDIAN__ 

}

/******************************************************************************
* check_central_func_init - check whether IXP1200 configured as central 
*  function, if it is, do some initialization 
*/
void check_central_func_init(void)
{
    ixp1200_chip_rev=(unsigned int)(*IXP1200_PCI_REV_CLASS & 0xFF);
    //DC put fllowing info to prevent running on chip with ver. before B0
/*
    if(((*IXP1200_PCI_REV_CLASS) & 0xFF) < 1)
    {
        while(1)
        {
    	    bsp_printf(" !!! Attention !!!\n");
    	    bsp_printf(" This Cygmon version can't run with IXP1200 chip ver. before B0\n");
    	    bsp_printf(" \n");
            ixp1200_sw_delay_msec(5000);   //DC delay about 5 sec
        }
    }
*/
#ifdef SHOW_VID_DID_REV 
    bsp_printf("IXP1200: VID=%4.4X, DID=%4.4X, CLASS=%8.8X, REV=%2.2X\n",
        ((*IXP1200_PCI_VEN_DEV_ID) & 0x0000FFFF), 
        (((*IXP1200_PCI_VEN_DEV_ID) & 0xFFFF0000) >> 16), 
        (((*IXP1200_PCI_REV_CLASS) & 0xFFFFFF00) >> 8),
        ((*IXP1200_PCI_REV_CLASS) & 0xFF));
#endif //#ifdef SHOW_VID_DID_REV 

#ifdef OUTPUT_DEBUG_INFO
    bsp_printf("In check_central_func(), before PCI reset SA_Control = %08X\n",
               *IXP1200_PCI_SA_CONTROL);
#endif //#ifdef OUTPUT_DEBUG_INFO

#ifdef OUTPUT_SA_CONTROL_DEBUG_INFO
    bsp_printf("In check_central_func(), before PCI reset SA_Control = %08X\n",
               *IXP1200_PCI_SA_CONTROL);
#endif //#ifdef OUTPUT_SA_CONTROL_DEBUG_INFO

    /* PCI reset */

    *IXP1200_PCI_SA_CONTROL = *IXP1200_PCI_SA_CONTROL & ~IXP1200_PCI_SA_CONTROL_PNR;

    delay_few_usecs();

#ifdef OUTPUT_DEBUG_INFO 
    bsp_printf("In check_central_func(), after PCI reset SA_Control = %08X\n",
               *IXP1200_PCI_SA_CONTROL);
#endif /* #ifdef OUTPUT_DEBUG_INFO */

    *IXP1200_PCI_CAP_PTR = 0xC;

    *IXP1200_PCI_DBELL_PCI_MASK = 0;

    *IXP1200_PCI_DBELL_SA_MASK = 0;

    *IXP1200_PCI_ADDR_EXT = 0x00000000;

    *IXP1200_PCI_SA_CONTROL = *IXP1200_PCI_SA_CONTROL | IXP1200_PCI_SA_CONTROL_PNR;

    delay_few_usecs();

    *IXP1200_PCI_DRAM_BASE_ADDR_MASK = IXP1200EB_SDRAM_SIZE - 0x00040000;

    *IXP1200_PCI_CSR_BASE_ADDR_MASK = 0x000c0000;

    *IXP1200_PCI_CMD_STAT= (*IXP1200_PCI_CMD_STAT & 0xFFFF0000);

    is_ixp1200_central_func = 
              IXP1200_PCI_SA_CONTROL_PCF0 & *IXP1200_PCI_SA_CONTROL;

    *IXP1200_PCI_SA_CONTROL =
        (*IXP1200_PCI_SA_CONTROL) | 0x00000001;

#ifdef OUTPUT_SA_CONTROL_DEBUG_INFO
    bsp_printf("Exit check_central_func_init(),  SA_Control = %08X\n",
               *IXP1200_PCI_SA_CONTROL);
#endif /* #ifdef OUTPUT_SA_CONTROL_DEBUG_INFO */
    return;

}

/******************************************************************************
* delay_few_usecs - wait a few usecs (no timers yet)
*/
static void delay_few_usecs(void)
{
    int m=0, n=0;
    for (m = 0; m < 60 ; m++)
        n = m * n + 3;
    return;
}

/******************************************************************************
* assign_pdev_addr - assign pci address resources for one pci device base address 
* register at location at offset
*/
static void assign_pdev_addr(struct pci_dev *pdev, int offset, int type, unsigned int size)
{
    unsigned int start_addr=0;
    
    if( type == IO_MAP)
    {
	start_addr=next_avl_io_addr;   
        next_avl_io_addr += size;   
    }
    else
    {
	start_addr=next_avl_mem_addr;
        next_avl_mem_addr += size;
    }
#ifdef DEBUG_PRINT_PCI_CONFIG_INFO
    bsp_printf(" in assign_pdev_addr: offset=%d, type=%d, size=%08X, start_addr=%08X\n", offset, type,size, start_addr);
    bsp_printf (" in assign_pdev_addr: next_avl_mem_sddr=%08X, next_avl_io_addr=%08X\n", next_avl_mem_addr, next_avl_io_addr);
#endif /* #ifdef DEBUG_PRINT_PCI_CONFIG_INFO */
    pci_write_config_dword(pdev, PCI_BASE_ADDRESS_0 + (offset*4), start_addr);
}


