/* System services for applications running on CygMon
 *
 * Copyright (c) 1997, 1998, 1999 Cygnus Solutions
 *
 * The authors hereby grant permission to use, copy, modify, distribute,
 * and license this software and its documentation for any purpose, provided
 * that existing copyright notices are retained in all copies and that this
 * notice is included verbatim in any distributions. No written agreement,
 * license, or royalty fee is required for any of the authorized uses.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 */
#include <unistd.h>
#include <monitor.h>
#include <signal.h>

#ifdef HAS_TIMER
#include <sys/time.h>
#endif

#if USE_CYGMON_PROTOTYPES
/* Use common prototypes */
/* Some of the composed board.h files compose these
   prototypes redundently, but if they dont,
   these are the common definitions */
#include "fmt_util.h"   /* Interface to string formatting utilities */
#include "tservice.h"   /* Interface to target specific services */
#include "generic-stub.h" /* from libstub */
#endif

#ifndef NULL
#define NULL 0
#endif


#ifdef HAS_TIMER
tick_type start_tickcount = 0;

int
gettimeofday (struct timeval *tp, struct timezone *junk)
{
  tick_type ticks = __read_ticks ();
  int hz = __get_timer_hz ();
  tp->tv_sec = ticks / hz;
  tp->tv_usec = (1000000 / hz) * (ticks % hz);
  return 0;
}

int
settimeofday (const struct timeval *tp, const struct timezone *junk)
{
  tick_type ticks;
  int hz = __get_timer_hz ();
  ticks = (tick_type)tp->tv_sec * hz + ((tick_type)tp->tv_usec * hz) / 1000000;
  __set_tick_counter (ticks);
  return 0;
}

static int
our_stime (long seconds)
{
  tick_type ticks = ((tick_type)seconds) * __get_timer_hz ();
  __set_tick_counter (ticks);
  return 0;
}
#endif


#ifdef CYGMON_SYSTEM_SERVICES

static int
userwrite (int des, target_register_t addr, int amt)
{
  int x;
  char ch;
  mem_addr_t mem;
  int res;

  if (amt < 0)
    {
      return -1;
    }
  if (des != 1 && des != 2)
    {
      return -1;
    }
  MAKE_STD_ADDR (addr, &mem);
  for (x = 0; x < amt; x++)
    {
      read_memory (&mem, 1, 1, &ch);
      if (des == 1 || des == 2)
	{
	  xprintf ("%c", ch);
	}
      else
	{
	  res = write (des, &ch, 1);
	  if (res < 0)
	    {
	      return res;
	    }
	}
      ADD_OFFSET (&mem, &mem, 1);
    }
  return amt;
}


static int
userread (int des, target_register_t addr, int amt)
{
  int x;
  char ch;
  mem_addr_t mem;
  int res;

  if (amt < 0)
    {
      return -1;
    }
  if (des != 0)
    {
      return -1;
    }
  MAKE_STD_ADDR (addr, &mem);
  for (x = 0; x < amt; x++)
    {
      if (des == 0)
	ch = xgetchar ();
      else
	{
	  res = read (des, &ch, 1);
	  if (res <= 0)
	    {
	      return res;
	    }
	}
      write_memory (&mem, 1, 1, &ch);
      ADD_OFFSET (&mem, &mem, 1);
    }
  return amt;
}




static int
do_exit ()
{
#ifdef HAS_TIMER
  tick_type tickend = __read_ticks ();
#endif
  clear_user_state ();
  if (stub_is_active)
    __send_exit_status (get_syscall_arg (0));
  else
    {
      xprintf ("*** EXIT code %d\n", get_syscall_arg (0));

#ifdef HAS_TIMER
      if (tickend != start_tickcount)
	{
	  long hz = __get_timer_hz ();
	  long tickdelta = tickend - start_tickcount;
	  long seconds = tickdelta / hz;
	  long dec = (tickdelta % hz) * 10000 / hz;

	  xprintf ("%d.%d seconds elapsed time, %d hz timer resolution\n",
		   seconds, dec, hz);
	}
#endif
    }
  return 1;
}


int
process_syscall (int which)
{
#ifdef TARGET_SYSCALLS
  {
    int res = target_syscalls (which);
    if (res)
      return res - 1;
  }
#endif
  switch (which) 
    {
    case 2: /* put one character to the tty; this is really "fork" */
      xprintf ("%c", get_syscall_arg (0));
      set_syscall_value (0);
      break;

    case 1: /* exit */
      return do_exit ();
      break;

    case 3: /* read */
      if (stub_is_active)
	set_syscall_value (__get_gdb_input (get_syscall_arg (1),
					    get_syscall_arg (2),
					    1));
      else
	set_syscall_value (userread (get_syscall_arg (0),
				     get_syscall_arg (1),
				     get_syscall_arg (2)));
      break;
    case 4: /* write */
      if (stub_is_active)
	set_syscall_value (__output_gdb_string (get_syscall_arg (1), 
						get_syscall_arg (2)));
      else
	set_syscall_value (userwrite (get_syscall_arg (0),
				      get_syscall_arg (1),
				      get_syscall_arg (2)));
      break;
    case 37: /* kill */
      {
	int sigval = get_syscall_arg (1);
	set_syscall_value (0);
	return -sigval;
      }

    case 48: /* signal */
      if (get_syscall_arg (0) == 0)
	{
	  user_signal_handler = (__PFI) get_syscall_arg (1);
	  set_syscall_value (0);
	}
      else
	set_syscall_value (-1);
      break;

#ifdef HAS_TIMER
    case 25: /* stime */
      {
	long *ptr = (long *)get_syscall_arg (0);
	set_syscall_value (our_stime (*ptr));
      }
      break;

    case 156: /* gettimeofday */
      set_syscall_value (gettimeofday ((struct timeval *)get_syscall_arg (0),
				       NULL));
      break;

    case 157: /* getitimer */
      set_syscall_value (getitimer (get_syscall_arg (0), 
				    (struct itimerval *)get_syscall_arg (1)));
      break;

    case 158: /* setitimer */
      set_syscall_value (setitimer (get_syscall_arg (0),
				    (struct itimerval *)get_syscall_arg (1),
				    (struct itimerval *)get_syscall_arg (2)));
      break;
#endif

#ifdef HAS_TIME
    case 13: /* time */
      set_syscall_value (time ((time_t *)get_syscall_arg (0)));
      break;
#endif

    case 184: /* get program arguments */
      set_syscall_value ((target_register_t)__get_program_args (get_syscall_arg (0)));
      break;

    default:
      xprintf ("Unknown syscall %d\n", which);
      return 1;
    }
  return 0;
}

#endif /* CYGMON_SYSTEM_SERVICES */

